/******************************************************************************
** Copyright (c) 2005
**	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
**
** ============================================================================
** Oct 13, 2005		Alex Feldman	Initial version.
**
******************************************************************************/

/******************************************************************************
**			   INCLUDE FILES
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <syslog.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/un.h>
#include <netinet/in.h>
#if defined(__LINUX__)
# include <linux/if.h>
# include <linux/if_packet.h>
#endif

#if defined(__LINUX__)
# include <linux/wanpipe_defines.h>
//# include <linux/wanpipe_common.h>
//# include <linux/if_wanpipe.h>
# include <linux/wanpipe_cfg.h>
#else
# include <net/wanpipe_defines.h>
# include <net/wanpipe_cfg.h>
//# include <net/wanpipe_common.h>
#endif

#include <wanec_iface.h>

/******************************************************************************
**			  DEFINES AND MACROS
******************************************************************************/
#define WAN_EC_DIR	"/etc/wanpipe/wan_ec"
#define WAN_EC_BUFFERS	WAN_EC_DIR "/buffers"

#define WAN_EC_DEBUG	"wan_ec_sangoma"

#define OCT6116_32S_IMAGE_NAME	"OCT6116-32S.ima"
#define OCT6126_128S_IMAGE_NAME	"OCT6126-128S.ima"
#define OCT6116_256S_IMAGE_NAME	"OCT6116-256S.ima"
#define OCT6116_32S_IMAGE_PATH	WAN_EC_DIR "/" OCT6116_32S_IMAGE_NAME
#define OCT6126_128S_IMAGE_PATH	WAN_EC_DIR "/" OCT6126_128S_IMAGE_NAME
#define OCT6116_256S_IMAGE_PATH	WAN_EC_DIR "/" OCT6116_256S_IMAGE_NAME

#define MAX_OPEN_RETRY	10

/******************************************************************************
**			STRUCTURES AND TYPEDEFS
******************************************************************************/

/******************************************************************************
**			   GLOBAL VARIABLES
******************************************************************************/
struct wan_ec_image
{
	int	hwec_chan_no;
	UINT32	max_channels;
	UINT32	memory_chip_size;
	UINT32	debug_data_mode;
	char	*image;	
	char	*image_path;	
} wan_ec_image_list[] =
	{
		{ 	16,
			32, cOCT6100_MEMORY_CHIP_SIZE_8MB,
			cOCT6100_DEBUG_GET_DATA_MODE_16S,
			OCT6116_32S_IMAGE_NAME, OCT6116_32S_IMAGE_PATH },
		{ 	32,
			32, cOCT6100_MEMORY_CHIP_SIZE_8MB,
			cOCT6100_DEBUG_GET_DATA_MODE_16S,
			OCT6116_32S_IMAGE_NAME, OCT6116_32S_IMAGE_PATH },
		{	128,
			128, cOCT6100_MEMORY_CHIP_SIZE_16MB,
			cOCT6100_DEBUG_GET_DATA_MODE_120S,
			OCT6126_128S_IMAGE_NAME, OCT6126_128S_IMAGE_PATH },
		{	256,
			256, cOCT6100_MEMORY_CHIP_SIZE_16MB,
			cOCT6100_DEBUG_GET_DATA_MODE_16S,
			OCT6116_256S_IMAGE_NAME, OCT6116_256S_IMAGE_PATH },
		{	0, 0, 0, 0, NULL }
	};
	
CHAR *ToneBufferPaths[WAN_NUM_PLAYOUT_TONES] = 
{
	WAN_EC_BUFFERS "/DTMF_0_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_1_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_2_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_3_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_4_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_5_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_6_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_7_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_8_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_9_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_A_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_B_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_C_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_D_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_STAR_ulaw.pcm",
	WAN_EC_BUFFERS "/DTMF_POUND_ulaw.pcm"
};
/******************************************************************************
** 			FUNCTION PROTOTYPES
******************************************************************************/
static int wanec_api_lib_open(wan_ec_api_t *ec_api);
static int wanec_api_lib_close(wan_ec_api_t *ec_api, int dev);
static int wanec_api_lib_ioctl(int dev, wan_ec_api_t *ec_api, int verbose);

/******************************************************************************
** 			FUNCTION DEFINITIONS
******************************************************************************/
static int wanec_api_lib_open(wan_ec_api_t *ec_api)
{
	int	dev;
	int	cnt = 0;

dev_open_again:

	dev = open(WANEC_DEV_DIR WANEC_DEV_NAME, O_RDONLY);
	if (dev < 0){
		if (cnt < MAX_OPEN_RETRY){
			cnt++;
			sleep(1);
			goto dev_open_again;
		}
		printf("ERROR: Failed to open device %s%s (err=%d,%s)\n",
					WANEC_DEV_DIR, WANEC_DEV_NAME,
					errno, strerror(errno));
		return -EINVAL;
	}
	return dev;
}

static int wanec_api_lib_close(wan_ec_api_t *ec_api, int dev)
{
	if (dev < 0){
		return -EINVAL;
	}  	
	close(dev);
	return 0;
}

static int wanec_api_lib_ioctl(int dev, wan_ec_api_t *ec_api, int verbose)
{
	int err;
	
	ec_api->err = WAN_EC_API_RC_OK;
	err = ioctl(dev, ec_api->cmd, ec_api);
	if (err){
		printf("Failed (%s)!\n", strerror(errno));
		return err;
	}
	if (ec_api->err){
		switch(ec_api->err){
		case WAN_EC_API_RC_INVALID_STATE:
			printf("Failed (Invalid State:%s)!\n",
					WAN_OCT6100_STATE_DECODE(ec_api->state));
			break;
		case WAN_EC_API_RC_FAILED:
		case WAN_EC_API_RC_INVALID_CMD:
		case WAN_EC_API_RC_INVALID_DEV:
		case WAN_EC_API_RC_BUSY:
		default:
			printf("Failed (%s)!\n",
					WAN_EC_API_RC_DECODE(ec_api->err));
			break;
		}
		return -EINVAL;
	}
	return 0;
}



/*
**	WAN EC Config
**/
static int
//wan_ec_loadImageFile(wan_ec_api_t *ec_api, struct wan_ec_image *image_info)
wanec_api_lib_loadImageFile(	wan_ec_api_t 	*ec_api,
				char		*path,
				PUINT8		*pData,
				UINT32		*size)
{
	UINT32	ulNumBytesRead;
	FILE*	pFile;
	PUINT8	ptr;
	
	/* Validate the parameters.*/
	if (path == NULL){
		printf("ERROR: %s: Invalid image filename!\n",
					ec_api->devname);
		return -EINVAL;
	}

	/*==================================================================*/
	/* Open the file.*/
	
	pFile = fopen(path, "rb" );
	if ( pFile == NULL ){
		printf("ERROR: %s: Failed to open image file %s!\n",
					ec_api->devname, path);
		return -EINVAL;
	}

	/*====================================================================*/
	/* Extract the file length.*/

	fseek( pFile, 0, SEEK_END );
	*size = (UINT32)ftell( pFile );
	fseek( pFile, 0, SEEK_SET );
	
	if (*size == 0 ){
		fclose( pFile );
		printf("ERROR: %s: Image file is too short!\n",
					ec_api->devname);
		return -EINVAL;
	}

	/*==================================================================*/
	/* Allocate enough memory to store the file content.*/
	
	ptr = (PUINT8)malloc(*size * sizeof(UINT8));
	if (ptr == NULL ){
		fclose( pFile );
		printf("ERROR: %s: Failed allocate memory for image file!\n",
					ec_api->devname);
		return -EINVAL;
	}

	/*==================================================================*/
	/* Read the content of the file.*/

	ulNumBytesRead = fread(	ptr,
				sizeof(UINT8),
				*size,
				pFile );
	if ( ulNumBytesRead != *size ){
		free( ptr );
		fclose( pFile );
		printf("ERROR: %s: Failed to read image file!\n",
					ec_api->devname);
		return -EINVAL;
	}

	/* The content of the file should now be in the f_pbyFileData memory,
	 * we can close the file and return to the calling function.*/
	fclose( pFile );
	*pData = ptr;	
	return 0;
}

/* Oct6100 image path */
static struct wan_ec_image *wanec_api_lib_image_search(int hwec_chan_no)
{
	struct wan_ec_image *image_info = &wan_ec_image_list[0];

	while(image_info && image_info->hwec_chan_no){
		if (image_info->hwec_chan_no == hwec_chan_no){
			return image_info;
		}
		image_info++;
	}	
	return NULL;
}

int wanec_api_lib_config(wan_ec_api_t *ec_api, int verbose)
{
	int			dev, err;
	struct wan_ec_image	*image_info;
	int			cnt = 0;

	printf("%s: Configuring Echo Canceller device...\t",
				ec_api->devname);

	dev = wanec_api_lib_open(ec_api);
	if (dev < 0){
		printf("Failed (Device open)!\n");
		return -ENXIO;
	}

	ec_api->cmd = WAN_EC_CMD_GETINFO;
	err = wanec_api_lib_ioctl(dev, ec_api, verbose);
	if (err || ec_api->err){
		wanec_api_lib_close(ec_api, dev);
		return (err) ? -EINVAL : 0;
	}
	
	image_info = wanec_api_lib_image_search(ec_api->u_info.max_channels);
	if (image_info == NULL){
		printf("Failed!\n\n");
		printf(
		"ERROR: %s: Failed to find image file to EC chip (max=%d)!\n",
					ec_api->devname,
					ec_api->u_info.max_channels);
		wanec_api_lib_close(ec_api, dev);
		return -EINVAL;
	}
	ec_api->u_config.max_channels		= ec_api->u_info.max_channels;
	ec_api->u_config.memory_chip_size	= image_info->memory_chip_size;
	ec_api->u_config.debug_data_mode	= image_info->debug_data_mode;
	
	/* Load the image file */
	err = wanec_api_lib_loadImageFile(	ec_api,
						image_info->image_path,
						&ec_api->u_config.imageData,
						&ec_api->u_config.imageSize);
	if (err){
		wanec_api_lib_close(ec_api, dev);
		return -EINVAL;
	}
	ec_api->cmd = WAN_EC_CMD_CONFIG;
	err = wanec_api_lib_ioctl(dev, ec_api, verbose);
	if (err || ec_api->err){
		free(ec_api->u_config.imageData);
		wanec_api_lib_close(ec_api, dev);
		return (err) ? -EINVAL : 0;
	}
	free(ec_api->u_config.imageData);
config_poll:
	ec_api->cmd = WAN_EC_CMD_GETINFO;
	err = wanec_api_lib_ioctl(dev, ec_api, verbose);
	if (err || ec_api->err){
		wanec_api_lib_close(ec_api, dev);
		return (err) ? -EINVAL : 0;
	}
	if (ec_api->state == WAN_OCT6100_STATE_CHIP_OPEN_PENDING){
		sleep(5);
		printf(".");
		if (cnt++ < 10) goto config_poll;
	}
	if (ec_api->state != WAN_OCT6100_STATE_CHIP_OPEN &&
	    ec_api->state != WAN_OCT6100_STATE_CHIP_READY){
		printf("Timeout!\n");
		wanec_api_lib_close(ec_api, dev);
		return -EINVAL;
	}

	/* Open all channels */	
	ec_api->cmd = WAN_EC_CMD_CHANNEL_OPEN;
	err = wanec_api_lib_ioctl(dev, ec_api, verbose);
	if (err || ec_api->err){
		wanec_api_lib_close(ec_api, dev);
		return (err) ? -EINVAL : 0;
	}
	
	printf("Done!\n");
	wanec_api_lib_close(ec_api, dev);
	return err;
}


int wanec_api_lib_toneload(wan_ec_api_t *ec_api)
{	
	UINT8	tone_path[100];
	int dev, err = 0;

	printf("%s: Start Tone loading...\t",
			ec_api->devname);
	dev = wanec_api_lib_open(ec_api);
	if (dev < 0){
		printf("Failed!\n");
		return -ENXIO;
	}

	sprintf(tone_path, "%s/%s.pcm", WAN_EC_BUFFERS, ec_api->u_tone_config.tone);
	err = wanec_api_lib_loadImageFile(	ec_api,
						tone_path,
						&ec_api->u_tone_config.data,
						&ec_api->u_tone_config.size);
	if (err){
		printf("Failed!\n");
		goto tone_load_done;
	}	
	
	err = wanec_api_lib_ioctl(dev, ec_api, ec_api->verbose);
	if (err || ec_api->err){
		goto tone_load_done;
	}	
	printf("Done (Index=%d)!\n", ec_api->u_tone_config.buffer_index);
	
tone_load_done:
	wanec_api_lib_close(ec_api, dev);
	if (ec_api->u_tone_config.data){
		free(ec_api->u_tone_config.data);
	}

	return err;
}

int wanec_api_lib_monitor(wan_ec_api_t *ec_api)
{
	int	dev, err;

	if (ec_api->channel){

		dev = wanec_api_lib_open(ec_api);
		if (dev < 0){
			return ENXIO;
		}

		printf("%s: Start monitorting the channel %d ...",
				ec_api->devname,
				ec_api->channel);
		err = wanec_api_lib_ioctl(dev, ec_api, 1);
		if (err || ec_api->err){
			wanec_api_lib_close(ec_api, dev);
			return (err) ? -EINVAL : 0;
		}
		printf("Done!\n");
		wanec_api_lib_close(ec_api, dev);
	}else{

		FILE		*output = NULL;
		size_t		len;
		int		cnt = 0;
		unsigned char	filename[MAXPATHLEN];
		
		printf("%s: Reading Monitor Data ..",
					ec_api->devname);
		do {
			dev = wanec_api_lib_open(ec_api);
			if (dev < 0){
				return ENXIO;
			}
			ec_api->channel = 0;
			ec_api->u_chan_monitor.remain_len = 0;
			ec_api->u_chan_monitor.remain_len = 0;
			ec_api->u_chan_monitor.data_len = 0;
			ec_api->u_chan_monitor.max_len = MAX_MONITOR_DATA_LEN;
			err = wanec_api_lib_ioctl(dev, ec_api, 0);
			if (err || ec_api->err){
				err = (err) ? -EINVAL : 0;
				goto monitor_done;
			}
			if (output == NULL){
				struct timeval	tv;
				struct tm	t;
				/* Firsr data */
				memset(filename, 0, MAXPATHLEN);
				gettimeofday(&tv, NULL);
				localtime_r((time_t*)&tv.tv_sec, &t);
				snprintf(filename, MAXPATHLEN,
				"%s_chan%d_%d.%d.%d_%d.%d.%d.bin",
					WAN_EC_DEBUG,
					ec_api->channel,
					t.tm_mon,t.tm_mday,1900+t.tm_year,
					t.tm_hour,t.tm_min,t.tm_sec);
				output = fopen(filename, "wb");
				if (output == NULL){
					printf("Failed!\n");
					printf("ERROR: %s: Failed to open binary file (%s)\n",
							ec_api->devname, filename);
					err = -EINVAL;
					goto monitor_done;
				}
			}
			len = fwrite(&ec_api->u_chan_monitor.data[0], sizeof(UINT8),
					ec_api->u_chan_monitor.data_len, output);
			if (len != ec_api->u_chan_monitor.data_len){
				printf("Failed!\n");
				printf("ERROR: %s: Failed to write to binary file (%s)!\n",
							ec_api->devname, filename);
				err = -EINVAL;
				goto monitor_done;
			}
			if (++cnt % 100 == 0){
				printf(".");
			}
			wanec_api_lib_close(ec_api, dev);
		}while(ec_api->u_chan_monitor.remain_len);
		if (output) fclose(output);
		printf("Done!\n");
		printf("Binary dump information is stored in %s file.\n", filename);
	}
monitor_done:
	return err;
}

int wanec_api_lib_cmd(wan_ec_api_t *ec_api)
{
	int	dev, err;
	
	dev = wanec_api_lib_open(ec_api);
	if (dev < 0){
		return -ENXIO;
	}
	printf("%s: Running %s command to Echo Canceller device...\t",
			ec_api->devname,
			WAN_EC_CMD_DECODE(ec_api->cmd));
	err = wanec_api_lib_ioctl(dev, ec_api, 1);
	if (err || ec_api->err){
		wanec_api_lib_close(ec_api, dev);
		return (err) ? -EINVAL : 0;
	}
	printf("Done!\n");
	wanec_api_lib_close(ec_api, dev);
	return 0;
}
